#include <string>
#include <vector>
#include <map>
#include <set>
#include <sstream>

using namespace std;

static bool hpt_initialized = false;

static map<string, int> hpt_dir;
static map<string, set<string> > hpt_child;

static void make_hpt()
{
  static string hpt[] = {
    "ADJP",   "right", "% QP JJ VBN VBG ADJP $ JJR JJS DT FW **** RBR RBS RB",
    "ADVP",   "left",  "RBR RB RBS FW ADVP CD **** JJR JJS JJ",
    "CONJP",  "left",  "CC RB IN", 
    "FRAG",   "left",  "**",
    "INTJ",   "right", "**", 
    "LST",    "left",  "LS :",
    "NAC",    "right", "NN NNS NNP NNPS NP NAC EX $ CD QP PRP VBG JJ JJS JJR ADJP FW",
    "NP",     "right", "EX $ CD QP PRP VBG JJ JJS JJR ADJP DT FW RB SYM PRP$",
    "NP$",    "right", "NN NNS NNP NNPS NP NAC EX $ CD QP PRP VBG JJ JJS JJR ADJP FW SYM",
    "PNP",    "right", "**",
    "PP",     "left",  "IN TO FW", 
    "PRN",    "left",  "**",
    "PRT",    "left",  "RP",
    "QP",     "right", "CD NCD % QP JJ JJR JJS DT", 
    "RRC",    "left",  "VP NP ADVP ADJP PP",
    "S",      "right", "VP SBAR ADJP UCP NP", 
    "SBAR",   "right", "S SQ SINV SBAR FRAG X",
    "SBARQ",  "right", "SQ S SINV SBARQ FRAG X", 
    "SINV",   "right", "S VP VBZ VBD VBP VB SINV ADJP NP",
    "SQ",     "right", "VP VBZ VBD VBP VB MD SQ", 
    "UCP",    "left",  "**", 
    "VP",     "left",  "VBD VBN MD VBZ TO VB VP VBG VBP ADJP NP",
    "WHADJP", "right", "JJ ADJP", 
    "WHADVP", "left",  "WRB",
    "WHNP",   "right", "WDT WP WP$ WHADJP WHPP WHNP",
    "WHPP",   "left",  "IN TO FW",
    "X",      "left",  "**",
    "///", "///", "///",
  };

  for (int i = 0;; i+=3) {
    const string lhs = hpt[i];
    const string dir = hpt[i+1];
    const string rhs = hpt[i+2];
    if (lhs == "///") break;
    hpt_dir[lhs] = (dir == "left") ? 1 : 0;
    istringstream is(rhs);
    string symbol;
    while (is >> symbol) {
      if (symbol == "**") continue;
      hpt_child[lhs].insert(symbol);
    }
  }
  
  hpt_initialized = true;
}

int percolate_head(const string & lhs, const vector<string> & rhs)
{
  if (rhs.size() == 1) return 0;
  if (!hpt_initialized) make_hpt();

  if (lhs == "NP") {
    for (int i = rhs.size() - 1; i >= 0; i--) {
      if (rhs[i][0] == 'N') return i;
    }
  }
  
  int dir = hpt_dir[lhs];
  set<string> & child = hpt_child[lhs];
  if (dir == 1) {
    for (int i = 0; i < rhs.size(); i++) {
      if (child.find(rhs[i]) != child.end()) return i;
    }
    return 0;
  }
  assert(dir == 0);
  if (dir == 0) {
    for (int i = rhs.size() - 1; i >= 0; i--) {
      if (child.find(rhs[i]) != child.end()) return i;
    }
    return rhs.size()-1;
  }
}
